/*
 * Copyright 2013 Texas Instruments, Inc.
 * Author: Dan Murphy <dmurphy@ti.com>
 *
 * Derived work from the pca953x.c driver
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <dm.h>
#include <i2c.h>
#include <tca642x.h>

/* tca642x register address definitions */
struct tca642x_bank_info tca642x_regs[] = {
	{ .input_reg = 0x00,
	  .output_reg = 0x04,
	  .polarity_reg = 0x08,
	  .configuration_reg = 0x0c },
	{ .input_reg = 0x01,
	  .output_reg = 0x05,
	  .polarity_reg = 0x09,
	  .configuration_reg = 0x0d },
	{ .input_reg = 0x02,
	  .output_reg = 0x06,
	  .polarity_reg = 0x0a,
	  .configuration_reg = 0x0e },
};

#define OFFSET_TO_BANK(x)			((x) >> 3) /* div by 8 === div by 2^3 === >> 3*/
#define OFFSET_TO_SHIFT(x)			((x) - ((OFFSET_TO_BANK(x)) << 3))
#define TCA642X_INPUT_REG(x)			((OFFSET_TO_BANK(x))+0x00)
#define TCA642X_OUTPUT_REG(x)		((OFFSET_TO_BANK(x))+0x04)
#define TCA642X_POLARITY_REG(x)		((OFFSET_TO_BANK(x))+0x08)
#define TCA642X_CONFIG_REG(x)		((OFFSET_TO_BANK(x))+0x0C)
#define NGPIOS	24

/*
 * Modify masked bits in register
 */
static int tca642x_reg_write(uchar chip, uint8_t addr,
		uint8_t reg_bit, uint8_t data)
{
	uint8_t valw;
	int org_bus_num;
	int ret;

	org_bus_num = i2c_get_bus_num();
	i2c_set_bus_num(CONFIG_SYS_I2C_TCA642X_BUS_NUM);

	if (i2c_read(chip, addr, 1, (uint8_t *)&valw, 1)) {
		printf("Could not read before writing\n");
		ret = -1;
		goto error;
	}
	valw &= ~reg_bit;
	valw |= data;

	ret = i2c_write(chip, addr, 1, (u8 *)&valw, 1);

error:
	i2c_set_bus_num(org_bus_num);
	return ret;
}

static int tca642x_reg_read(uchar chip, uint8_t addr, uint8_t *data)
{
	uint8_t valw;
	int org_bus_num;
	int ret = 0;

	org_bus_num = i2c_get_bus_num();
	i2c_set_bus_num(CONFIG_SYS_I2C_TCA642X_BUS_NUM);
	if (i2c_read(chip, addr, 1, (u8 *)&valw, 1)) {
		ret = -1;
		goto error;
	}

	*data = valw;

error:
	i2c_set_bus_num(org_bus_num);
	return ret;
}

/*
 * Set output value of IO pins in 'reg_bit' to corresponding value in 'data'
 * 0 = low, 1 = high
 */
int tca642x_set_val(uchar chip, uint8_t gpio_bank,
					uint8_t reg_bit, uint8_t data)
{
	uint8_t out_reg = tca642x_regs[gpio_bank].output_reg;

	return tca642x_reg_write(chip, out_reg, reg_bit, data);
}

int tca642x_set_value_state(struct tca642x_state * state, unsigned pin, unsigned value)
{
	debug("in %s(pin=%d, value=%d\n", __func__, pin, value);
	state->bank[OFFSET_TO_BANK(pin)].output_reg &= ~(0x01 << OFFSET_TO_SHIFT(pin));
	state->bank[OFFSET_TO_BANK(pin)].output_reg |= (value  << OFFSET_TO_SHIFT(pin));
	return 0;
}
int tca642x_set_value(uchar chip, unsigned pin, unsigned value)
{
	return tca642x_reg_write(chip, TCA642X_OUTPUT_REG(pin), (1 << OFFSET_TO_SHIFT(pin)), (value << OFFSET_TO_SHIFT(pin)));
}
int tca642x_set_polarity(uchar chip, unsigned pin, unsigned polarity)
{
	return tca642x_reg_write(chip, TCA642X_POLARITY_REG(pin),
									(1 << OFFSET_TO_SHIFT(pin)),
									(((uint8_t)(polarity & 0xFF))<< OFFSET_TO_SHIFT(pin)));
}
/*
 * Set read polarity of IO pins in 'reg_bit' to corresponding value in 'data'
 * 0 = read pin value, 1 = read inverted pin value
 */
int tca642x_set_pol(uchar chip, uint8_t gpio_bank,
					uint8_t reg_bit, uint8_t data)
{
	uint8_t pol_reg = tca642x_regs[gpio_bank].polarity_reg;

	return tca642x_reg_write(chip, pol_reg, reg_bit, data);
}

/*
 * Set direction of IO pins in 'reg_bit' to corresponding value in 'data'
 * 0 = output, 1 = input
 */
int tca642x_set_dir(uchar chip, uint8_t gpio_bank,
					uint8_t reg_bit, uint8_t data)
{
	uint8_t config_reg = tca642x_regs[gpio_bank].configuration_reg;

	return tca642x_reg_write(chip, config_reg, reg_bit, data);
}

int tca642x_set_direction(uchar chip, unsigned pin, unsigned in) /* 1 for input, 0 for output */
{
	debug("%s: pin %u is on bank 0x%0X, bit 0x%02X\n", __func__, pin, OFFSET_TO_BANK(pin), OFFSET_TO_SHIFT(pin));
	int  res = tca642x_reg_write(chip, TCA642X_CONFIG_REG(pin), (1 << OFFSET_TO_SHIFT(pin)), (in << OFFSET_TO_SHIFT(pin)));
	return res;
}
int tca642x_set_direction_input_state(struct tca642x_state * state, unsigned pin)
{
	debug("in %s(pin=%d \n", __func__, pin);
	state->bank[OFFSET_TO_BANK(pin)].configuration_reg |= (1 << OFFSET_TO_SHIFT(pin));
	return 0;
}
int tca642x_set_direction_input(uchar chip, unsigned pin)
{
	return tca642x_set_direction(chip, pin, 1);
}

int tca642x_set_direction_output_state(struct tca642x_state * state, unsigned pin, unsigned value)
{
	tca642x_set_value_state(state,pin, value);
	state->bank[OFFSET_TO_BANK(pin)].configuration_reg &= ~(1 << OFFSET_TO_SHIFT(pin));
	return 0;
}
int tca642x_set_direction_output(uchar chip, unsigned pin, unsigned value)
{
	tca642x_set_value(chip, pin, value);
	return tca642x_set_direction(chip, pin, 0);
}


/*
 * Read current logic level of all IO pins
 */
int tca642x_get_val(uchar chip, uint8_t gpio_bank)
{
	uint8_t val;
	uint8_t in_reg = tca642x_regs[gpio_bank].input_reg;

	if (tca642x_reg_read(chip, in_reg, &val) < 0)
		return -1;

	return (int)val;
}

int tca642x_get_value(uchar chip, unsigned pin)
{
	debug("%s: pin %u is on bank 0x%0X, bit 0x%02X\n", __func__, pin, OFFSET_TO_BANK(pin), OFFSET_TO_SHIFT(pin));
	uint8_t val = 0;
	int  res = tca642x_reg_read(chip, TCA642X_INPUT_REG(pin), &val);
	if (res < 0)
		return res;
	return !!(res && (1 << OFFSET_TO_SHIFT(pin)));
}

void tca642x_print_state(struct tca642x_state * state)
{
	for(int i = 0; i < 3; i++){
		printf("BANK %d:\n", i);
		printf("\tCONFIG:   0x%02X\n", state->bank[i].configuration_reg);
		printf("\tOUTPUT:   0x%02X\n", state->bank[i].output_reg);
		printf("\tINPUT:    0x%02X\n", state->bank[i].input_reg);
		printf("\tPOLARITY: 0x%02X\n", state->bank[i].polarity_reg);

	}
}
int tca642x_write_state(struct udevice * dev, struct tca642x_state * state)
{
	int i, ret=0;
	uchar chip = ((struct dm_i2c_chip*)dev_get_parent_platdata(dev))->chip_addr;
	debug("Got chip 0x%02X\n", chip);

	for (i = 0; i < 3; i++) {
		debug("Writing 0x%02X to tca642x config register %02X\n", state->bank[i].configuration_reg, TCA642X_CONFIG_REG(i << 3));
		ret = tca642x_reg_write(chip,TCA642X_CONFIG_REG(i << 3), 0xFF, state->bank[i].configuration_reg);
		if (ret) { break; }
		debug("Writing 0x%02X to tca642x out register %02X\n", state->bank[i].output_reg, TCA642X_OUTPUT_REG(i << 3));
		ret = tca642x_reg_write(chip,TCA642X_OUTPUT_REG(i << 3), 0xFF, state->bank[i].output_reg);
		if (ret) { break; }
		debug("Writing 0x%02X to tca642x pol register %02X\n", state->bank[i].polarity_reg, TCA642X_POLARITY_REG(i << 3));
		ret = tca642x_reg_write(chip,TCA642X_POLARITY_REG(i << 3),0xFF, state->bank[i].polarity_reg);
		if (ret) { break; }
	}
	if (ret) {
		debug("%s has failed to read state\n", __func__);
		return ret;
	}
	return 0;
}
int tca642x_read_state(struct udevice * dev, struct tca642x_state *state)
{
	int i, ret=0;
	uchar chip = ((struct dm_i2c_chip*)dev_get_parent_platdata(dev))->chip_addr;

	for (i = 0; i < 3; i++) {
		ret = tca642x_reg_read(chip,TCA642X_CONFIG_REG(i << 3), &state->bank[i].configuration_reg);
		debug("Reading 0x%02X to tca642x config register %02X\n", state->bank[i].configuration_reg, TCA642X_CONFIG_REG(i << 3));
		if (ret) { break; }

		ret = tca642x_reg_read(chip,TCA642X_OUTPUT_REG(i << 3), &state->bank[i].output_reg);
		debug("Reading 0x%02X to tca642x out register %02X\n", state->bank[i].output_reg, TCA642X_OUTPUT_REG(i << 3));
		if (ret) { break; }

		ret = tca642x_reg_read(chip,TCA642X_POLARITY_REG(i << 3), &state->bank[i].polarity_reg);
		debug("Reading 0x%02X to tca642x pol register %02X\n", state->bank[i].polarity_reg, TCA642X_POLARITY_REG(i << 3));
		if (ret) { break; }
	}
	if (ret) {
		printf("%s has failed to read state\n", __func__);
		return ret;
	}
	return 0;
}
/*
 * Set the inital register states for the tca642x gpio expander
 */
int tca642x_set_inital_state(uchar chip, struct tca642x_bank_info init_data[])
{
		
	int i, ret;
	uint8_t config_reg;
	uint8_t polarity_reg;
	uint8_t output_reg;

	for (i = 0; i < 3; i++) {
		config_reg = tca642x_regs[i].configuration_reg;
		ret = tca642x_reg_write(chip, config_reg, 0xff,
				init_data[i].configuration_reg);
		polarity_reg = tca642x_regs[i].polarity_reg;
		ret = tca642x_reg_write(chip, polarity_reg, 0xff,
				init_data[i].polarity_reg);
		output_reg = tca642x_regs[i].output_reg;
		ret = tca642x_reg_write(chip, output_reg, 0xff,
				init_data[i].output_reg);
	}

	return ret;
}

#ifdef CONFIG_CMD_TCA642X
/*
 * Display tca642x information
 */
static int tca642x_info(uchar chip)
{
	int i, j;
	uint8_t data;

	printf("tca642x@ 0x%x (%d pins):\n", chip, 24);
	for (i = 0; i < 3; i++) {
		printf("Bank %i\n", i);
		if (tca642x_reg_read(chip,
				     tca642x_regs[i].configuration_reg,
				     &data) < 0)
			return -1;
		printf("\tConfiguration: ");
		for (j = 7; j >= 0; j--)
			printf("%c", data & (1 << j) ? 'i' : 'o');
		printf("\n");

		if (tca642x_reg_read(chip,
				     tca642x_regs[i].polarity_reg, &data) < 0)
			return -1;
		printf("\tPolarity: ");
		for (j = 7; j >= 0; j--)
			printf("%c", data & (1 << j) ? '1' : '0');
		printf("\n");

		if (tca642x_reg_read(chip,
				     tca642x_regs[i].input_reg, &data) < 0)
			return -1;
		printf("\tInput value: ");
		for (j = 7; j >= 0; j--)
			printf("%c", data & (1 << j) ? '1' : '0');
		printf("\n");

		if (tca642x_reg_read(chip,
				     tca642x_regs[i].output_reg, &data) < 0)
			return -1;
		printf("\tOutput value: ");
		for (j = 7; j >= 0; j--)
			printf("%c", data & (1 << j) ? '1' : '0');
		printf("\n");
	}

	return 0;
}

cmd_tbl_t cmd_tca642x[] = {
	U_BOOT_CMD_MKENT(device, 3, 0, (void *)TCA642X_CMD_DEVICE, "", ""),
	U_BOOT_CMD_MKENT(output, 4, 0, (void *)TCA642X_CMD_OUTPUT, "", ""),
	U_BOOT_CMD_MKENT(input, 3, 0, (void *)TCA642X_CMD_INPUT, "", ""),
	U_BOOT_CMD_MKENT(invert, 4, 0, (void *)TCA642X_CMD_INVERT, "", ""),
	U_BOOT_CMD_MKENT(read, 3, 0, (void *)TCA642X_CMD_READ, "", ""),
	U_BOOT_CMD_MKENT(info, 2, 0, (void *)TCA642X_CMD_INFO, "", ""),
};

int do_tca642x(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	static uchar chip = CONFIG_SYS_I2C_TCA642X_ADDR;
	int ret = CMD_RET_USAGE ;
	uint8_t gpio_bank = 0;
	uint8_t bank_shift;
	ulong ul_arg2 = 0;
	ulong ul_arg3 = 0;
	cmd_tbl_t *c;

	c = find_cmd_tbl(argv[1], cmd_tca642x, ARRAY_SIZE(cmd_tca642x));

	/* All commands but "device" require 'maxargs' arguments */
	if (!c ||
	    !((argc == (c->maxargs)) ||
	    (((int)c->cmd == TCA642X_CMD_DEVICE) &&
	    (argc == (c->maxargs - 1))))) {
		return CMD_RET_USAGE;
	}

	/* arg2 used as chip number or pin number */
	if (argc > 2)
		ul_arg2 = simple_strtoul(argv[2], NULL, 10);

	/* arg3 used as pin or invert value */
	if (argc > 3) {
		ul_arg3 = simple_strtoul(argv[3], NULL, 10) & 0x1;
	}

	switch ((int)c->cmd) {
	case TCA642X_CMD_INFO:
		ret = tca642x_info(chip);
		if (ret)
			ret = CMD_RET_FAILURE;
		break;

	case TCA642X_CMD_DEVICE:
		if (argc == 3)
			chip = (uint8_t)ul_arg2;
		printf("Current device address: 0x%x\n", chip);
		ret = CMD_RET_SUCCESS;
		break;

	case TCA642X_CMD_READ:
		ret = tca642x_get_value(chip, ul_arg2);
		if (ret < 0){
			ret = CMD_RET_FAILURE;
			break;
		}
		printf("chip 0x%02x, pin %lu = %d\n", chip, ul_arg2, ret);
	case TCA642X_CMD_INPUT:
		ret = tca642x_set_direction_input(chip, ul_arg2);
		if (ret)
			ret = CMD_RET_FAILURE;
		break;

	case TCA642X_CMD_OUTPUT:
		ret = tca642x_set_direction_output(chip, ul_arg2, ul_arg3 ? 0 : 1);
		if (ret)
			ret = CMD_RET_FAILURE;
		break;

	case TCA642X_CMD_INVERT:
		bank_shift = ul_arg2 - (gpio_bank * 10);
		ret = tca642x_set_pol(chip, gpio_bank, (1 << bank_shift), (ul_arg3 << bank_shift));
		if (ret)
			ret = CMD_RET_FAILURE;
		break;
	}

	if (ret == CMD_RET_FAILURE)
		eprintf("Error talking to chip at 0x%x\n", chip);

	return ret;
}

U_BOOT_CMD(
	tca642x,	5,	1,	do_tca642x,
	"tca642x gpio access",
	"device [dev]\n"
	"	- show or set current device address\n"
	"tca642x info\n"
	"	- display info for current chip\n"
	"tca642x output pin 0|1\n"
	"	- set pin as output and drive low or high\n"
	"tca642x invert pin 0|1\n"
	"	- disable/enable polarity inversion for reads\n"
	"tca642x input pin\n"
	"	- set pin as input and read value"
);

#endif /* CONFIG_CMD_TCA642X */
